/***********************************************************
 * $Id: DataCodingHelper.java 258 2010-04-18 19:22:38Z util $
 * 
 * optional of the clazzes.org project
 * 
 * http://www.clazzes.org
 *
 * Created: 09.04.2010, wglas
 *
 ***********************************************************/

package org.clazzes.optional.io;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Static helper methods, which complement the functionality of
 * {@link DataInput} and {@link DataOutput}.
 */
public abstract class DataCodingHelper {

    /**
     * Write a three-byte unsigned integer value to the stream.
     * @param out The stream to write to.
     * @param x The value, of which the 24 low-order bits are written.
     * @throws IOException Upon I/O errors.
     */
    public static void writeUInt3(DataOutput out, int x) throws IOException {
        out.write(x>>>16);
        out.writeShort(x);
    }
  
    /**
     * Read a three-byte unsigned integer from a stream.
     * @param in The stream to be read.
     * @return A value between 0 and 16777215.
     * @throws IOException Upon errors or premature end of stream.
     */
    public static int readUInt3(DataInput in) throws IOException {
        
        return in.readUnsignedByte()<<16 | in.readUnsignedShort();
    }

    /**
     * Read a four-byte unsigned integer from a stream.
     * @param in The stream to be read.
     * @return An unsigned integer value represented as a long value.
     * @throws IOException Upon errors or premature end of stream.
     */
    public static long readUInt4(DataInput in) throws IOException {
        
        return ((long)in.readInt()) & 0xffffffffL;
    }

    /**
     * Read a dynamic opaque byte array from the stream. The number of bytes read as length
     * prefix is determined from the given maximal size.
     * 
     * Maximal sizes from 0 to 255 read one byte of length information,
     * maximal sizes from 256 to 65535 read two bytes of length information,
     * maximal sizes from 65536 to 16777215 read three bytes of length information,
     * maximal sizes greater or equal to 16777216 or negative maximal sizes read
     * four bytes of length information.
     * 
     * @param in The stream to read from.
     * @param maxSize The maximal permissible size or a negative number for no size restrictions.
     * @return A newly allocated byte array with a size as specified in the length prefix
     *         inside the stream. If the length in the stream is 0, <code>null</code> is returned.
     * @throws IOException Upon read errors.
     */
    public static byte[] readDynamicOpaque(DataInput in, int maxSize) throws IOException {
        
        int length = readArrayLength(in, maxSize);
        
        if (length <= 0) return null;
        
        byte [] ret = new byte[length];
        in.readFully(ret);
        return ret;
    }
    
    private static int readArrayLength(DataInput in, int maxSize) throws IOException {
        int length = 0;
        if (maxSize >= 0 && maxSize < 256)
            length = in.readUnsignedByte();
        else if (maxSize >= 0 && maxSize < 65536)
            length = in.readUnsignedShort();
        else if (maxSize >= 0 && maxSize < 16777216)
            length = readUInt3(in);
        else
            length = in.readInt();
            
        if (length < 0 || (maxSize >= 0 && length > maxSize))
            throw new IOException("Size of array filed exceeds maximal allowed size.");
        
        return length;
    }
    
    private static void writeArrayLength(DataOutput out, int length, int maxSize) throws IOException{
        if (maxSize >= 0 && length > maxSize)
            throw new IOException("Size of array filed exceeds maximal allowed size.");

        if (maxSize >= 0 && maxSize < 256)
            out.write(length);
        else if (maxSize >= 0 && maxSize < 65536)
            out.writeShort(length);
        else if (maxSize >= 0 && maxSize < 16777216)
            writeUInt3(out,length);
        else
            out.writeInt(length);
    }
    
    /**
     * Write a dynamic opaque byte array to the stream. The number of bytes written as length
     * prefix is determined from the given maximal size.
     * 
     * Maximal sizes from 0 to 255 write one byte of length information,
     * maximal sizes from 256 to 65535 write two bytes of length information,
     * maximal sizes from 65536 to 16777215 write three bytes of length information,
     * maximal sizes greater or equal to 16777216 or negative maximal sizes write
     * four bytes of length information.
     * 
     * @param out The stream to write to.
     * @param maxSize The maximal permissible size or a negative number for no size restrictions.
     * @param opaque The byte array to write. If <code>null</code> is passed in, an array of
     *               length 0 written.
     * @throws IOException Upon write errors.
     */
    public static void writeDynamicOpaque(DataOutput out, int maxSize, byte[] opaque) throws IOException {
        
        if (opaque == null)
            writeArrayLength(out,0, maxSize);
        else {
            writeArrayLength(out, opaque.length, maxSize);
            out.write(opaque);
        }
    }

    /**
     * Read a dynamic integer array from the stream. The number of bytes read as length
     * prefix is determined from the given maximal size.
     * 
     * Maximal sizes from 0 to 255 read one byte of length information,
     * maximal sizes from 256 to 65535 read two bytes of length information,
     * maximal sizes from 65536 to 16777215 read three bytes of length information,
     * maximal sizes greater or equal to 16777216 or negative maximal sizes read
     * four bytes of length information.
     * 
     * @param in The stream to read from.
     * @param maxSize The maximal permissible size or a negative number for no size restrictions.
     * @return A newly allocated <code>int[]</code> with a size as specified in the length prefix
     *         inside the stream. If the length in the stream is 0, <code>null</code> is returned.
     * @throws IOException Upon read errors.
     */
    public static int[] readDynamicIntArray(DataInput in, int maxSize) throws IOException {
        
        int length = readArrayLength(in, maxSize);
        
        if (length <= 0) return null;
        
        int [] ret = new int[length];
        for (int i=0; i<length; i++)
            ret[i] = in.readInt();
        return ret;
    }
    
    /**
     * Write a dynamic integer array to the stream. The number of bytes written as length
     * prefix is determined from the given maximal size.
     * 
     * Maximal sizes from 0 to 255 write one byte of length information,
     * maximal sizes from 256 to 65535 write two bytes of length information,
     * maximal sizes from 65536 to 16777215 write three bytes of length information,
     * maximal sizes greater or equal to 16777216 or negative maximal sizes write
     * four bytes of length information.
     * 
     * @param out The stream to write to.
     * @param maxSize The maximal permissible size or a negative number for no size restrictions.
     * @param arr The <code>int[]</code> to write. If <code>null</code> is passed in, an array of
     *               length 0 written.
     * @throws IOException Upon write errors.
     */
    public static void writeDynamicIntArray(DataOutput out, int maxSize, int[] arr) throws IOException {
        int n = arr == null ? 0 : arr.length;
        writeArrayLength(out, n, maxSize);
        for (int i=0; i<n; i++)
            out.writeInt(arr[i]);
    }
}
