/***********************************************************
 * $Id: BigInt.java 253 2010-04-16 08:07:16Z util $
 * 
 * Tiny TLS for Java ME.
 * http://www.clazzes.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Created: 11.12.2009
 * 
 ***********************************************************/

package org.clazzes.optional.lang;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import org.clazzes.optional.io.DataCodeable;
import org.clazzes.optional.io.DataCodingHelper;
import org.clazzes.optional.io.Serializable;

/**
 * A variable length integer class.
 */
public final class BigInt implements Serializable, DataCodeable {

    private static final long serialVersionUID = 8753764488075517746L;

    /**
     * Either 1 for a positive number of -1 for a negative number.
     */
    private final int sign;
    
    /**
     * The number of non-null bytes in {@link BigInt#v}.
     * {@link BigInt#length} is at least 1.
     */
    private final int length;
    
    /**
     * The value of this BigInt stored in LSB-first format.
     */
    private final int[] v;
    
    private static final long INT32MASK = 0xffffffffL;
    
    private static int b2i(byte b) {
        return ((int)b)&0xff;
    }
    
    private static long i2l(int i) {
        return ((long)i)&INT32MASK;
    }
    
    private static int findLength(int [] v) {
        int i = v.length;
        while (i > 1) {
            if (v[i-1] != 0) break;
            --i;
        }
        return i;
    }
    
    /**
     * Construct a string from an internal representation. Using this constructor may break
     * the the immutability of the generated instance, when the passed array is modified
     * after calling this constructor.
     * 
     * This constructor is however very useful, when initializing static instances in
     * a memory optimized version.
     * 
     * @param sign The sign of the returned instance. Should be <code>1</code> or <code>-1</code>.
     * @param v The internal integer array starting with the least significant 32 bits.
     * 
     * @see #toJavaInitializer(String)
     */
    public BigInt(int sign, int[] v) {
        super();
        this.length = findLength(v);
        // zero is positive by definition.
        this.sign = (this.length == 1 && v[0] == 0) ? 1 : (sign < 0 ? -1 : 1);
        this.v = v;
    }

    private BigInt(int sign, int[] v, int length) {
        super();
        this.length = length;
        // zero is positive by definition.
        this.sign = this.length == 1 && v[0] == 0  ? 1 : sign;
        this.v = v;
    }

    public BigInt(int i) {
        this.sign = i<0 ? -1 : 1;
        this.v = new int[]{i<0 ? -i : i};
        this.length = 1;
    }
    
    public BigInt(long i) {
        this.sign = i<0 ? -1 : 1;
        if (i<0) i=-i;
        this.v = new int[]{(int)i,(int)(i>>32)};
        this.length = (this.v[1] == 0 ? 1 : 2);
    }
    
    /**
     * Create a big integer from a MSB raw value.
     * 
     * @param sign If negative, this number is negative.
     * @param raw An array of raw bytes starting with the most significant one.
     */
    public BigInt(int sign, byte [] raw) {
        this.sign = sign < 0 ? -1 : 1;
        
        if (raw.length == 0) {
            this.v=new int[1];
            this.length = 1;
        }
        else {
            this.v = new int[(raw.length+3)/4];
            
            int j = raw.length % 4;
            int iv = this.v.length;
            
            switch(j) {
            case 1:
                this.v[--iv] = b2i(raw[0]);
                break;
            case 2:
                this.v[--iv] = b2i(raw[0])<<8 | b2i(raw[1]);
                break;
            case 3:
                this.v[--iv] = b2i(raw[0])<<16 | b2i(raw[1])<<8 | b2i(raw[2]);
                break;
            }
            
            for (int i=j;i<raw.length;i+=4) {
                this.v[--iv] = b2i(raw[i])<<24 | b2i(raw[i+1])<<16 | b2i(raw[i+2])<<8 | b2i(raw[i+3]);
            }
            this.length = findLength(this.v);
        }
    }
    
//#ifndef JAVAME
    /**
     * Perform <code>v=fac*v+x</code> in place an return the new length of <code>v</code>.
     * @param v The number to multiply.
     * @param length The number of significant int's in <code>v</code>
     * @param fac The positive multiplicand.
     * @param x The positive number to add.
     * @return The new number of significant int's in <code>v</code>.
     */
    static private int multiplyAddLong (int[] v, int length, long fac, long x) {
        
        long tmp = x;
        
        for (int i=0;i<length;++i) {
            tmp += i2l(v[i]) * fac;
            v[i] = (int)tmp;
            tmp >>>= 32;
        }
        
        if (tmp != 0) {
            v[length] = (int)tmp;
            ++length;
        }
        return length;
    }
    
    static private int digit2Value(char c) {
        
        if ('0' <= c && c <= '9')
            return c-'0';
        
        throw new NumberFormatException("Invalid digit ["+c+"] in integral number.");
    }
//#endif
    
    /**
     * Parse the given decimal string and return the resulting BigInt.
     * @param s A string consisting of an optional <code>-</code> sign followed
     *          by a sequence of digits <code>0...9</code>.
     * @return A BigInt representing the parsed sequence.
     */
//#ifndef JAVAME
    public static BigInt valueOf(String s) {
        
        int sign = 1;
        int i=0;
        if (s.charAt(i) == '-') { ++i; sign = -1; }
        
        // The truth is: multiply with ln(10)/ln(2^32) = .10381025296523007337
        // division by 9 is a good starting point and overestimates the number of bytes.
        // However, any floating point arithmetic may be too much for Java ME... 
        int[] v = new int[(s.length()-i+8)/9];
        int l = 1;
        
        v[0] = digit2Value(s.charAt(i));
        ++i;
        
        for(;i<s.length();++i) {
            
            l=multiplyAddLong(v,l,10L,digit2Value(s.charAt(i)));
        }
        
        return new BigInt(sign,v);
    }
//#endif
    
//#ifndef JAVAME
    static private int hexDigit2Value(char c) {
        
        if ('0' <= c && c <= '9')
            return c-'0';
        
        if ('A' <= c && c <= 'F')
            return 10+c-'A';
        
        if ('a' <= c && c <= 'f')
            return 10+c-'a';
       
        throw new NumberFormatException("Invalid hexadecimal digit ["+c+"] in integral number.");
    }
//#endif
    
   /**
     * Parse the given hexadecimal string and return the resulting BigInt.
     * @param s A string consisting of an optional <code>-</code> sign followed
     *          by a sequence of digits <code>0...9,A...F</code>.
     * @return A BigInt representing the parsed sequence.
     */
  //#ifndef JAVAME
    public static BigInt hexValueOf(String s) {
        
        
        int sign = 1;
        int i=0;
        if (s.charAt(i) == '-') { ++i; sign = -1; }
        
        int[] v = new int[(s.length()-i+7)/8];

        int l = v.length-1;
        
        switch ((s.length()-i) % 8) {
        
        case 7:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 24;
        case 6:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 20;
        case 5:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 16;
        case 4:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 12;
        case 3:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 8;
        case 2:
            v[l] |= hexDigit2Value(s.charAt(i++)) << 4;
        case 1:
            v[l] |= hexDigit2Value(s.charAt(i++));
            --l;
        }
        
        while(i<s.length()) {
            
            v[l] |= hexDigit2Value(s.charAt(i++)) << 28;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 24;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 20;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 16;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 12;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 8;
            v[l] |= hexDigit2Value(s.charAt(i++)) << 4;
            v[l] |= hexDigit2Value(s.charAt(i++));
            --l;
        }
        
        return new BigInt(sign,v);
    }
  //#endif
    
    /**
     * Parse an ASN.1 INTEGER representation. This representation
     * is in MSB-first two's complement format with the following restriction,
     * that the bits of the first octet and the first bit of the second octet
     * may not all be zero or one. 
     * 
     * @param asn1 The content octets of an ASN.1 INTEGER. 
     * @return A new BigInt instance.
     */
//#ifndef JAVAME
    public static BigInt valueOf(byte[] asn1) {
        
        if (asn1.length <=0 || asn1[0] >= 0)
            return new BigInt(1,asn1);
        
        int [] v = new int[(asn1.length+3)/4];
        
        int j = asn1.length % 4;
        int iv = v.length;
        
        switch(j) {
        case 1:
            v[--iv] = b2i(asn1[0]) | 0xffffff00;
            break;
        case 2:
            v[--iv] = b2i(asn1[0])<<8 | b2i(asn1[1]) | 0xffff0000;
            break;
        case 3:
            v[--iv] = b2i(asn1[0])<<16 | b2i(asn1[1])<<8 | b2i(asn1[2]) | 0xff000000;
            break;
        }
        
        for (int i=j;i<asn1.length;i+=4) {
            v[--iv] = b2i(asn1[i])<<24 | b2i(asn1[i+1])<<16 | b2i(asn1[i+2])<<8 | b2i(asn1[i+3]);
        }
        
        long tmp = 0L;
        
        for (iv=0;iv<v.length;++iv) {
            tmp -= i2l(v[iv]);
            v[iv] = (int)tmp;
            tmp >>= 32;
        }
        
       return new BigInt(-1,v);
    }
//#endif
    
    /**
     * Read a positive BigInt from the given
     * data input using a two byte length prefix.
     * 
     * @param in The input stream to read from.
     */
    public static BigInt readDataInstance(DataInput in) throws IOException {
        
        int length = in.readUnsignedShort();
        
        if (length == 0) return null;  
        
        int[] rep = new int[(length+3)/4];
        
        int j=rep.length;
        
        switch (length%4) {
        case 3:
            rep[--j] = DataCodingHelper.readUInt3(in);
            break;
        case 2:
            rep[--j] = in.readUnsignedShort();
            break;
        case 1:
            rep[--j] = in.readUnsignedByte();
            break;
        }

        while (j > 0) {
            rep[--j] = in.readInt();
        }
        
        return new BigInt(1,rep);
    }
    
    
    /**
     * Write the MSB magnitude of this BigInt to the given
     * data output using a two byte length prefix.
     * 
     * @param out The output stream to write to.
     */
    public void writeData(DataOutput out) throws IOException {
        
        int i = this.length;
     
        int xx = this.v[--i];
         
        if (xx < 0 || xx > 0xffffff) {
             
            out.writeShort(4*i+4);
            out.writeInt(xx);
        }
        else if (xx > 0xffff) {
            
            out.writeShort(4*i+3);
            DataCodingHelper.writeUInt3(out,xx);
        }
        else if (xx > 0xff) {
            
            out.writeShort(4*i+2);
            out.writeShort(xx);
        }
        else {
            
            out.writeShort(4*i+1);
            out.write(xx);
        }
        
        while (i>0) {
            xx = this.v[--i];
            out.writeInt(xx);
        }     
    }

    /**
     * @return The length of the unsigned magnitude as returned by {@link #getMsbMagnitude()}
     *         in bytes.
     */
    public int getByteLength() {
        int i = this.length;
        
        int xx = this.v[--i];
        
        if (xx < 0 || xx > 0xffffff) {
            return 4*i+4;
        }
        else if (xx > 0xffff) {
            return 4*i+3;
        }
        else if (xx > 0xff) {
            return 4*i+2;
        }
        else {
            return 4*i+1;
        }
    }
    
    /**
     * @return The raw MSB magnitude of this integer.
     */
    public byte[] getMsbMagnitude() {
        return this.getMsbMagnitude1(false);
    }
    
    private byte[] getMsbMagnitude1(boolean asn1) {
                   
        int i=0;
        int j;
        int iv=this.length;
        --iv;

        // in asn1 mode account for an additional zero byte if most significant bit is set.
        if (this.v[iv] < 0) {
            j=0;
            if (asn1) ++i;
            
        } else if (this.v[iv] <= 0xff) {
            j = -3;
            if (asn1 && this.v[iv] > 0x7f) ++i;
        }
        else if (this.v[iv] <= 0xffff) {
            j = -2;
            if (asn1 && this.v[iv] > 0x7fff) ++i;
        }
        else if (this.v[iv] <= 0xffffff) {
            j = -1;
            if (asn1 && this.v[iv] > 0x7fffff) ++i;
        }
        else {
            j = 0;
        }
            
        byte[] ret= new byte[4*this.length+j+i];
        
        switch (j) {
        case -1:
            ret[i++] = (byte)(this.v[iv] >>> 16); 
        case -2:
            ret[i++] = (byte)(this.v[iv] >>> 8); 
        case -3:
            ret[i++] = (byte)(this.v[iv]);
            --iv;
            break;
        }
        
        
        for (;iv>=0;--iv) {
            
            ret[i++] = (byte)(this.v[iv] >>> 24); 
            ret[i++] = (byte)(this.v[iv] >>> 16); 
            ret[i++] = (byte)(this.v[iv] >>> 8); 
            ret[i++] = (byte)(this.v[iv]);
        }
        
        return ret;
    }
    
    /**
     * Convert this BigInt to an ASN1 MSB first integer representation.
     * 
     * @return A MSB first two's complement representation of this BigInt.
     */
//#ifndef JAVAME
    public byte[] getASN1() {
        
        if (this.sign >= 0)
            return this.getMsbMagnitude1(true);
        
        int i=0;
        int j;
        int iv=this.length;
        --iv;

        // in asn1 mode account for an additional zero byte if most significant bit is set.
        if (this.v[iv] < 0) {
            j = 0;
            if (this.v[iv] == 0x80000000) {
                int ii=0;
                for (ii=0;ii<iv&&this.v[ii] == 0;++ii);
                if (ii < iv) ++i;
            }
            else
                ++i;
        }
        else if (this.v[iv] <= 0xff) {
            j = -3;
            if (this.v[iv] > 0x80) ++i;
            else if (this.v[iv] == 0x80) {
                int ii=0;
                for (ii=0;ii<iv&&this.v[ii] == 0;++ii);
                if (ii < iv) ++i;                
            }
        }
        else if (this.v[iv] <= 0xffff) {
            j = -2;
            if (this.v[iv] > 0x8000) ++i;
            else if (this.v[iv] == 0x8000) {
                int ii=0;
                for (ii=0;ii<iv&&this.v[ii] == 0;++ii);
                if (ii < iv) ++i;                
            }
        }
        else if (this.v[iv] <= 0xffffff) {
            j = -1;
            if (this.v[iv] > 0x800000) ++i;
            else if (this.v[iv] == 0x800000) {
                int ii=0;
                for (ii=0;ii<iv&&this.v[ii] == 0;++ii);
                if (ii < iv) ++i;                
            }
        }
        else {
            j = 0;
        }
            
        byte[] ret= new byte[4*this.length+j+i];
        
        long tmp = 0L;
        i = ret.length;
        
        for (int ii=0;ii<iv;++ii) {
            
            tmp -= i2l(this.v[ii]);
            
            ret[--i] = (byte)(tmp);
            ret[--i] = (byte)(tmp >>> 8); 
            ret[--i] = (byte)(tmp >>> 16); 
            ret[--i] = (byte)(tmp >>> 24); 
            tmp>>=32;
        }
        
        tmp -= i2l(this.v[iv]);
        
        ret[--i] = (byte)(tmp);
        if (j >= -2) {
      
            ret[--i] = (byte)(tmp >>> 8); 
      
            if (j >= -1) {
                ret[--i] = (byte)(tmp >>> 16); 
                
                if (j>=0)
                    ret[--i] = (byte)(tmp >>> 24);
            }
        }
        
        // prepend 0xff where appropriate.
        if (i>0) ret[--i] = -1;

        return ret;
    }

    private static final char HEX_DIGITS[] = new char[] {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
    
    /**
     * Convert a BigInt to a hexadecimal string.
     * 
     * @return An uppercase hexadecimal string with an optional minus sign
     *         and an even number of hexadecimal, uppercase digits.
     */
    public String toHexString() {
        
        int j;
        int iv=this.length;
        --iv;
        
        if (this.v[iv] < 0 || this.v[iv] > 0xfffffff)
            j = 0;
        else if (this.v[iv] > 0xffffff)
            j = -1;
        else if (this.v[iv] > 0xfffff)
            j = -2;
        else if (this.v[iv] > 0xffff)
            j = -3;
        else if (this.v[iv] > 0xfff)
            j = -4;
        else if (this.v[iv] > 0xff)
            j = -5;
        else if (this.v[iv] > 0xf)
            j = -6;
        else
            j = -7;
        
        char ret[] = new char[8*this.length+j+(this.sign<0?1:0)];

        int i=0;
        
        if (this.sign < 0)
            ret[i++] = '-';
        
        switch (j) {
        case 0:
            ret[i++] = HEX_DIGITS[this.v[iv] >>> 28];
        case -1:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 24)&0x0f];
        case -2:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 20)&0x0f];
        case -3:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 16)&0x0f];
        case -4:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 12)&0x0f];
        case -5:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 8)&0x0f];
        case -6:
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 4)&0x0f];
        case -7:
           ret[i++] = HEX_DIGITS[this.v[iv]&0x0f];
        }

        --iv;
        
        for (;iv>=0;--iv) {
         
            ret[i++] = HEX_DIGITS[this.v[iv] >>> 28];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 24)&0x0f];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 20)&0x0f];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 16)&0x0f];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 12)&0x0f];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 8)&0x0f];
            ret[i++] = HEX_DIGITS[(this.v[iv] >>> 4)&0x0f];
            ret[i++] = HEX_DIGITS[this.v[iv]&0x0f];
        }
        
        return new String(ret);
    }
//#endif
    
    /**
     * @return The sign of this integer.
     */
    public int getSign() {
        return this.sign;
    }
    
    /**
     * Compare this number with <code>a</code> by taking signs into account.
     * @param a the number to compare this instance with.
     * @return <code>-1,0</code> or <code>1</code> if this number
     *         is less than, equal to or greater than <code>a</code>. 
     */
    public int compareTo(BigInt a) {
        
        if (this.sign != a.sign) return this.sign < 0 ? -1 : 1;
        if (this.length < a.length) return this.sign < 0 ? 1 : -1;
        if (this.length > a.length) return this.sign < 0 ? -1 : 1;

        int i=this.length;
        
        while (--i >= 0) {
            if (i2l(this.v[i]) < i2l(a.v[i])) return this.sign < 0 ? 1 : -1;
            if (i2l(this.v[i]) > i2l(a.v[i])) return this.sign > 0 ? -1 : 1;
        }
        
        return 0;
    }
    
    /**
     * Compare this number with <code>a</code> by taking signs into account.
     * @param a the number to compare this instance with.
     * @return <code>-1,0</code> or <code>1</code> if this number
     *         is less than, equal to or greater than <code>a</code>. 
     */
    public int compareTo(int a) {
        
        if (a>0 && this.sign < 0) return -1;
        if (a<0 && this.sign > 0) return 1;
        if (this.length > 1 || this.v[0] < 0) return this.sign > 0 ? 1 : -1;

        if (this.sign > 0) {
            return this.v[0] - a;
        }
        else {
            return a + this.v[0];
        }
    }
    
    /**
     * Compare this number with <code>a</code> by ignoring the sign.
     * @param a the number to compare this instance with.
     * @return <code>-1,0</code> or <code>1</code> if the magnitude of this number
     *         is less than, equal to or greater than <code>a</code>. 
     */
    private int compareToUnsigned(BigInt a) {
        
        if (this.length < a.length) return -1;
        if (this.length > a.length) return 1;

        int i=this.length;
        
        while (--i >= 0) {
            if (i2l(this.v[i]) < i2l(a.v[i])) return -1;
            if (i2l(this.v[i]) > i2l(a.v[i])) return 1;
        }
        
        return 0;
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object o) {
        
        if (o==this) return true;
        if (!(o instanceof BigInt)) return false;
        
        BigInt a = (BigInt)o;
        
        if (this.sign != a.sign) return false;
        if (this.length != a.length) return false;

        int i=this.length;
        
        while (--i >= 0) {
            if (this.v[i] != a.v[i]) return false;
        }
        
        return true;
        
    }
    
    /**
     * @return A decimal representation of this BigInt.
     */
//#ifndef JAVAME
    public String toString() {
        
        char[] ch = new char[this.length*10+1];
        
        int [] vv = new int[this.length];
        System.arraycopy(this.v,0,vv,0,this.length);
        
        int l = this.length;
        int nc = ch.length;
        
        do {
            
            long digit = i2l(vv[0])%10L;

            for (int i=1; i<l;++i) {
                
                // This make use of the fact, that (2^(4*x)) % 10 == 6 for all x >= 1.
                digit += 6L * (i2l(vv[i])%10L);
            }
            
            ch[--nc] = (char) ('0' + digit%10L);
            
            long rem = 0L;
            
            int i = l;
            while (--i >= 0) {
                
                long tmp = i2l(vv[i])+ rem * 0x100000000L;
                rem = tmp % 10L;
                vv[i] = (int)(tmp / 10L);
            }
                
            while (l>0 && vv[l-1] == 0) --l;
            
        } while (l>0);
        
        if (this.sign < 0)
            ch[--nc] = '-';
        
        return new String(ch,nc,ch.length-nc);
    }
//#endif
    
  //#ifndef JAVAME
    /**
     * Return a jString, which may be inserted in JAVA source code to
     * initialize a static instance in the form
     * 
     * <pre>
     *   static final BigInt identifier = new BigInt(sign,
     *     new int[] {
     *       0xdeabbeef,0xadbedba,....
     *     });
     * </pre>
     * 
     * @param identifier The name of the identifier to use.
     */
    public String toJavaInitializer(String identifier) {
        
        StringBuffer sb = new StringBuffer();
        
        sb.append("  static final BigInt ");
        sb.append(identifier);
        sb.append(" = new BigInt(");
        sb.append(this.sign);
        sb.append(",\n");
        sb.append("    new int[] {");
        
        for (int i=0;i<this.length;++i) {
            
            if (i==0)
                sb.append("\n      ");
            else if (i%8 == 0)
                sb.append(",\n      ");
            else
                sb.append(", ");
            
            sb.append("0x");
            sb.append(Integer.toHexString(this.v[i]));   
        }
        sb.append("\n    });\n");
      
        return sb.toString();
    }
  //#endif
    
    /**
     * Add <code>a</code> to <code>this</code> and return the result.
     * @param a The number to add.
     * @return <code>this+a</code>.
     */
//#ifndef JAVAME
    public BigInt add(BigInt a) {
        
        if(a.sign == this.sign)
            return new BigInt(this.sign,addUnsigned(this,a));
        else {
            
            int r = this.compareToUnsigned(a);
            
            if (r==0)
                return new BigInt(0);
           
            if (r>0)
                return new BigInt(this.sign,subUnsigned(this,a)); 
            else
                return new BigInt(a.sign,subUnsigned(a,this));
        }
    }
//#endif
    
    /**
     * Subtract <code>a</code> from <code>this</code> and return the result.
     * @param a The number to subtract.
     * @return <code>this-a</code>.
     */
//#ifndef JAVAME
   public BigInt sub(BigInt a) {
        
        if(a.sign != this.sign)
            return new BigInt(this.sign,addUnsigned(this,a));
        else {
            
            int r = this.compareToUnsigned(a);
            
            if (r==0)
                return new BigInt(0);
           
            if (r>0)
                return new BigInt(this.sign,subUnsigned(this,a)); 
            else
                return new BigInt(-this.sign,subUnsigned(a,this));
        }
    }
//#endif
   
   /**
     * Add the given numbers by ignoring the sign.
     * @param a The first operand
     * @param b The second operand.
     * @return The internal representation of the sum of both numbers.
     */
//#ifndef JAVAME
    private static int[] addUnsigned(BigInt a, BigInt b) {
                    
        int l = b.length > a.length ? b.length : a.length;
            
        if ((b.length >= a.length && b.v[b.length-1] < 0) ||
                (a.length >= b.length && a.v[a.length-1] < 0 ) )
            ++l;
        
        int[] v = new int[l];
        
        long tmp = 0L;
        int i=0;
        
        for (;i<b.length&&i<a.length;++i) {
            
            tmp += i2l(b.v[i]) + i2l(a.v[i]);
            v[i] = (int)tmp;
            
            tmp >>>= 32;
        }
        
        for (;i<b.length;++i) {
            
            tmp += i2l(b.v[i]);
            v[i] = (int)tmp;
            
            tmp >>>= 32;
        }
        
        for (;i<a.length;++i) {
            
            tmp += i2l(a.v[i]);
            v[i] = (int)tmp;
            
            tmp >>>= 32;
        }
       
        if (i<l) v[i] = (int)tmp;
        return v;
    }
//#endif
    
    /**
     * Subtract <code>b</code> from <code>a</code> by ignoring the sign and assuming
     * that <code>a</code> is larger than <code>b</code>.
     * 
     * @param a The first operand
     * @param b The second operand.
     * @return The internal representation of the sum of both numbers.
     */
//#ifndef JAVAME
    private static int[] subUnsigned(BigInt a, BigInt b) {
                    
        int[] v = new int[a.length];
        
        long tmp = 0L;
        int i=0;
        
        for (;i<b.length;++i) {
            
            tmp += i2l(a.v[i]);
            tmp -= i2l(b.v[i]);
            v[i] = (int)tmp;
            
            tmp >>= 32;
        }
        
        for (;i<a.length;++i) {
            
            tmp += i2l(a.v[i]);
            v[i] = (int)tmp;
            
            tmp >>= 32;
        }
        
        return v;
    }
//#endif
    
    /**
     * Multiply <code>a</code> and <code>this</code> and return the result.
     * @param a The number to multiply.
     * @return <code>this*a</code>.
     */
//#ifndef JAVAME
    public BigInt mul(BigInt a) {
        
        int[] v = new int[this.length+a.length];
       
        int v_l = mul0(v,this.v,this.length,a.v,a.length);
        return new BigInt(this.sign * a.sign,v,v_l);
    }
//#endif
    
    /**
     * Worker subroutine for multiplying two numbers without taking signs into account.
     * 
     * @param v The result with a length of at least <code>a_l+b_l</code>.
     * @param a The first multiplicand.
     * @param a_l The length of the first multiplicand.
     * @param b The second multiplicand.
     * @param b_l The length of the second multiplicand.
     * @return The length of the result.
     */
//#ifndef JAVAME
    private static int mul0(int[] v, int[] a, int a_l, int[] b,int b_l) {
        
        int ret = 1;
            
        long tmp = 0L;
        int j,k;
        
        for (int i=0;i<a_l+b_l;++i) {
            
            if (i < b_l) {
                j = 0;
                k = i;
            }
            else {
                k = b_l - 1;
                j = i-k;
            }

            long vi;
            
            if (tmp != 0L) {
                vi = tmp & INT32MASK;
                tmp>>>=32;
            }
            else {
                vi = 0L;
            }
            
            for (;j<a_l && k >= 0;++j,--k) {
                
                long t = i2l(a[j])*i2l(b[k]);
                
                vi += t & INT32MASK;
                tmp += t >>> 32;
            }
            tmp += vi >>> 32;
            v[i] = (int)vi;
            
            if (i>0 && v[i] != 0)
                ret = i+1;
        }
        return ret;
    }
//#endif

    /**
     * Calculate the highest bit set in the magnitude of this number.
     * The bit with value <code>1</code> is counted as bit <code>0</code>,
     * the bit with value <code>1&lt;&lt;x</code> is counted as bit <code>x</code>.
     * For a value of zero, <code>-1</code> is returned.
     * 
     * @return The number of the highest bit set, starting with <code>0</code>
     *         at the least significant bit.
     */
    public int getHighestBit() {
        
        return (this.length-1)*32 + highestBit(this.v[this.length-1]);
    }
    
    /**
     * Calculate the highest bit set in the given number.
     * The bit with value <code>1</code> is counted as bit <code>0</code>,
     * the bit with value <code>1&lt;&lt;x</code> is counted as bit <code>x</code>.
     * For an input value of zero, <code>-1</code> is returned.
     * 
     * @param i An arbitrary <code>int</code> value treated as unsigned.
     * @return The number of the highest bit set, starting with <code>0</code>
     *         at the least significant bit.
     */
    private static int highestBit(int i) {
        
        if (i<0) return 31;
        
        int n0 = 0;
        int n1 = 31;
        
        while (n1 > n0) {

            int n = n0 + (n1 -n0)/2;
            int mask = 1 << n;
            
            if (mask <= i) {
                n0 = n+1;
            } else {
                n1 = n;
            }
        }
        
        return n0-1;
    }
    
    /**
     * Calculate <code>a-=b&lt;&lt;&lt;nshift</code>. 
     * 
     * @param a The left operand.
     * @param al The length of <code>a</code>.
     * @param b The right operand.
     * @param bl The length of <code>b</code>.
     * @param nshift The number of bits to shift.
     * @return The resulting length of <code>a</code>.
     */
    private static int sub0shifted(int[] a, int al, int [] b, int bl, int nshift) {
        
        // number of byte to shift.
        int bshift = nshift>>>5;
        int i = 0;
        nshift &= 0x1F;
        
        long tmp = 0L;
        int ret = 0;
        
        for (;i<bshift;++i) {
            if (a[i] != 0)
                ret = i;
        }
        
        for (;i<bl+bshift;++i) {
           
            tmp += i2l(a[i]);
            tmp -= i2l(b[i-bshift])<<nshift;
            a[i] = (int)tmp;
            if (a[i] != 0)
                ret = i;

            tmp >>= 32;
        }
        
        for (;i<al;++i) {
            
            tmp += i2l(a[i]);
            a[i] = (int)tmp;
            if (a[i] != 0)
                ret = i;
           
            tmp >>= 32;
        }
        
        return ret+1;
    }
   
    private static int compare0shifted(int [] a, int al, int [] b, int bl, int nshift) {
        
        // number of byte to shift.
        int bshift = nshift>>>5;
        nshift &= 0x1F;

        if (nshift == 0) {
            if (al < bl + bshift) return -1; 
            if (al > bl + bshift) return 1;
        
            while (--al >= bshift) {
                if (i2l(a[al]) < i2l(b[al-bshift])) return -1;
                if (i2l(a[al]) > i2l(b[al-bshift])) return 1;
            }
            
        }
        else {
            
            if (al < bl + bshift) return -1; 
            if (al > bl + bshift+1) return 1;
            
            long tmp = i2l(b[bl-1]) << nshift;

            if (al == bl + bshift+1) {
                
                --al;
                if (i2l(a[al]) < tmp >>> 32) return -1;
                if (i2l(a[al]) > tmp >>> 32) return 1;
        
                
            }
            else {
                if (tmp>>>32 != 0) return -1;
            }
            tmp <<= 32;
            // al == bl+bshift
           
            while (--al > bshift) {
            
                tmp += i2l(b[al-bshift-1]) << nshift;
            
                if (i2l(a[al]) < tmp >>> 32) return -1;
                if (i2l(a[al]) > tmp >>> 32) return 1;
                tmp <<= 32;
            }
            
            if (al >= 0) {
                
                if (i2l(a[al]) < tmp >>> 32) return -1;
                if (i2l(a[al]) > tmp >>> 32) return 1;
                --al;
            }
        }
        
        while (al >= 0) {
            if(a[al] != 0) return 1;
            --al;
        }
            
        return 0;
    }
    
    /**
     * Worker subroutine for in-place calculation of the modulus of
     * a number.
     *  
     * @param a The number to calculate the modulus from.
     * @param l The length of <code>a</code>
     * @param b The denominator with respect to which to take the modulus. 
     * @return The length of the resulting number which is stored in <code>a</code>.
     */
    private static int mod0(int [] a, int l, BigInt b) {
        
        if (compare0shifted(a,l,b.v,b.length,0)<0)
            return l;
        
        int a_l = l;

        int b_hb = highestBit(b.v[b.length-1]);
        int a_hb = highestBit(a[l-1]);
        int bs_l;
        
        if (a_hb > b_hb) {
            bs_l = l+1;
            b_hb += 32;
        } else {
            bs_l=l;
        }

        int i = 32*(bs_l-b.length) - (b_hb - a_hb);

        while (i>=0) {
                        
            // a >= bs ?
            if (compare0shifted(a,a_l,b.v,b.length,i)>=0) {
                a_l = sub0shifted(a,a_l,b.v,b.length,i);
                if (a_l < 0) return 1;
            }

            --i;
        }
        
        return a_l;
    }
    
    /**
     * Calculate <code>this</code> modulo <code>b</code>.
     * @param b
     * @return <code>this % b</code>.
     */
    public BigInt mod(BigInt b) {
        
        int [] a = new int[this.length];
        System.arraycopy(this.v,0,a,0,this.length);
        int a_l = mod0(a,this.length,b);
        // This is the sign convention used by GNU bc
        return new BigInt(this.sign,a,a_l);
    }
    
    
    /**
     * Multiply <code>a</code> and <code>this</code> and return the result modulo
     * <code>b</code>.
     * @param a The number to multiply.
     * @param b The modulus.
     * @return <code>(this*a)%b</code>.
     */
//#ifndef JAVAME
    public BigInt mulMod(BigInt a, BigInt b) {
        
        int[] v = new int[this.length+a.length];
        int v_l = mul0(v,this.v,this.length,a.v,a.length);
        v_l = mod0(v,v_l,b);
        return new BigInt(this.sign*a.sign,v,v_l);
    }
//#endif
    
    /**
     * Calculate <code>this^n</code> and return the result modulo
     * <code>b</code>.
     * @param n The exponent.
     * @param b The modulus.
     * @return <code>(this^n)%b</code>.
     */
    public BigInt powMod(int n, BigInt b) {
       
        if (n<0)
            throw new IllegalArgumentException("Negative exponent.");
       
        if (n==0)
            return new BigInt(1);
       
        if (n==1)
            return this;
        
        int[] e = new int[]{n};
        if ((b.v[0] & 1) == 0)
//#ifdef JAVAME
//@         throw new IllegalArgumentException("Even moduli are not supported on JAVAME.");
//#else
            // slow exponentiation for even moduli.
            return powmod0(e,e.length,b);
//#endif
        else
            // fast montgomery exponentiation for odd moduli.
            return powmodmont0(e,e.length,b);
    }
    
    /**
     * Calculate <code>this^n</code> and return the result modulo
     * <code>b</code>.
     * @param n The exponent.
     * @param b The modulus.
     * @return <code>(this^n)%b</code>.
     */
    public BigInt powMod(BigInt n, BigInt b) {
       
        if (n.getSign()<0)
            throw new IllegalArgumentException("Negative exponent.");
       
        if (n.compareTo(0) == 0)
            return new BigInt(1);
       
        if (n.compareTo(1) == 0)
            return this;
        
        if ((b.v[0] & 1) == 0)
//#ifdef JAVAME
//@         throw new IllegalArgumentException("Even moduli are not supported on JAVAME.");
//#else
            // slow exponentiation for even moduli.
            return powmod0(n.v,n.length,b);
//#endif
        else
            // fast montgomery exponentiation for odd moduli.
            return powmodmont0(n.v,n.length,b);
    }
    
//#ifndef JAVAME
    private BigInt powmod0(int[] e, int el, BigInt b) {
        
        /* ai holds a^(2^ibit) mod b */
        int[] ai = new int[2*b.length];
        System.arraycopy(this.v,0,ai,0,this.length);
        int ai_l = mod0(ai,this.length,b);
        int[] v = new int[2*b.length];
        int v_l = -1;
       
        int[] tmp = new int[2*b.length];
       
        int ibit,mask;
        
        for (int i=0;i<el;++i) {
        
            int nbits = i < el-1 ? 32 : highestBit(e[i]);
        
            for (ibit=0,mask=1;ibit<32;ibit++,mask<<=1) {
            
                // is bit i set in n?
                if ((e[i]&mask) != 0) {
                
                    if (v_l<1) {
                    /* result is a^(2^ibit) mod b, if this is the first bit set. */
                    System.arraycopy(ai,0,v,0,ai_l);
                    v_l = ai_l;
                    }
                    else {
                        /* multiply result with b^(2^ibit) mod n */
                        v_l = mul0(tmp,v,v_l,ai,ai_l);
                        // swap tmp and v.
                        { int[] xxx=tmp; tmp=v; v=xxx; }
                        v_l = mod0(v,v_l,b);
                    }
                }

                /* any bits left ? */
                if (ibit >=nbits) break;
                
                // go on from a^(2^ibit) to a^(2^(ibit+1)) = (a^(2^ibit))^2
                ai_l = mul0(tmp,ai,ai_l,ai,ai_l);
                // swap tmp and ai.
                { int[] xxx=tmp; tmp=ai; ai=xxx; }
                ai_l = mod0(ai,ai_l,b);
            }
        }
        
        return new BigInt((e[0]&1)==0 ? 1 : this.sign,v,v_l);
    }
//#endif
    
    private static boolean isEven(long x) {
        
        return (x&1) == 0L;
    }

    /**
     * @param x the operand
     * @return <code>y</code>, such that <code>x*y mod 2^32 == 1</code>
     */
    private static long inverse32 (long x) {
        
        long y = 0x100000000L;
        
        if (isEven(x))
            throw new ArithmeticException("Even Numbers have no inverse modulo 2^32.");
        
        long u=x,A=1,B=0,C=0,D=1;

        for (int i=0;i<32;++i) {
        
            if (!isEven(C) || !isEven(D)) {
                C += y;
                D -= x;
            }
            C/=2;
            D/=2;
        }
      
        do {
            
            while (isEven(u)) {
                u>>>=1;
            
                if (!isEven(A) || !isEven(B)) {
                    A += y;
                    B -= x;
                }
                A/=2;
                B/=2;
            }
            
            --u;
            A-=C;
            B-=D;
           
        } while (u != 0L);
        
        return C;
    }

    /**
     * @param m The modulus.
     * @return m', such that <code>m' = -m^-1 mod 2^32</code> 
     */
    private static long calculateMM(BigInt m) {
        return (-inverse32(m.v[0]))&INT32MASK;
    }
    
    /**
     * Montgomery multiplication.
     * 
     * @param a The result vector, where <code>y*x*R^-1 mod m</code> is stored. A has to be
     *          allocated to a size of <code>m.length+1</code>.
     * @param x The first operand, which his assumed to have a length of <code>m.length</code>.
     * @param y The second operand, which has a length of <code>yl</code>.
     * @param m The modulus.
     * @param mm The result of <code>calculateMM(m)</code>
     * @return The length of the calculated result a, which is less than <code>m</code>. 
     */
    private int montmult0(int[] a, int[]x, int[] y, int yl, BigInt m, long mm) {

        /*
         * For the following we choose b := 2^32.
         * INPUT: integers m = (m_n-1 · · · m_1 m_0 )b ,
         *       x = (x_n-1 · · · x_1 _x0 )b , y = (y_n-1 · · · y_1 y_0 )b
         * with 0 ≤ x, y < m, R = b^n with gcd(m, b) = 1, and m' = -m^-1 mod b.
         * OUTPUT: xyR^-1 mod m.
         *    1. A←0. (Notation: A = (a_n a_n-1 · · · a_1 a_0 )b .)
         *    2. For i from 0 to (n − 1) do the following:
         *          2.1 ui ←(a0 + xi y0 )m' mod b.
         *          2.2 A←(A + xi y + ui m)/b.
         *    3. If A ≥ m then A←A - m.
         *    4. Return(A)
         */

        // i==0
        long ui = (((i2l(x[0]) * i2l(y[0]))&INT32MASK) * mm)&INT32MASK;
        long tmp = 0L,ttmp=0L;
        
        for (int j = 0;j<m.length;++j) {
            
            if (j<yl)
                ttmp += i2l(x[0]) * i2l(y[j]);
            tmp += ttmp >>> 32;
            ttmp &= INT32MASK;
            ttmp += ui * i2l(m.v[j]);
            tmp += ttmp >>> 32;
        
             if (j>0)
                a[j-1] = (int)ttmp;
             /*
               else
                  assert (ttmp & INT32MASK) == 0L;
             */
             
            ttmp = tmp & INT32MASK;
            tmp >>>= 32;
        }
        a[m.length-1] = (int)ttmp;
        a[m.length]   = (int)tmp;
        
        for (int i=1;i<m.length;++i) {
            
            ui = ((i2l(a[0]) + ((i2l(x[i]) * i2l(y[0]))&INT32MASK)) * mm)&INT32MASK;
            tmp = 0L;
            ttmp= 0L;
            
            for (int j = 0;j<m.length;++j) {
                
                ttmp += i2l(a[j]);
                if (j<yl)
                    ttmp += i2l(x[i]) * i2l(y[j]);
                
                tmp += ttmp >>> 32;
                ttmp &= INT32MASK;
                ttmp += ui * i2l(m.v[j]);
                tmp += ttmp >>> 32;
                
                if (j>0)
                    a[j-1] = (int)ttmp;
                /*
                 else
                    assert (ttmp & INT32MASK) == 0L;
                */
                
                ttmp = tmp & INT32MASK;
                tmp >>>= 32;
            }
            
            ttmp += i2l(a[m.length]);
            a[m.length-1] = (int)ttmp;
            tmp += ttmp >>> 32;
            a[m.length]   = (int)tmp;
        }
        
        
        int al = findLength(a);
        
        if (compare0shifted(a,al,m.v,m.length,0) >= 0) {
            al = sub0shifted(a,al,m.v,m.length,0);
        }
        return al;
    }
        
    private BigInt powmodmont0(int[] e, int el, BigInt m) {
        
        /*
          INPUT: m = (ml−1 · · · m0 )b , R = b^l , m = -m^-1 mod b, e = (et · · · e0)2 with et = 1,
          and an integer x, 1 ≤ x < m.
          OUTPUT: x^e mod m.
             1. x← Mont(x, R2 mod m), A←R mod m. (R mod m and R2 mod m may be pro-
                   vided as inputs.)
             2. For i from t down to 0 do the following:
                 2.1 A← Mont(A, A).
                 2.2 If ei = 1 then A← Mont(A, x).
             3. A← Mont(A, 1).
             4. Return(A).
         */
        
        long mm = calculateMM(m);
        
        // mm * m = -1 mod 2^32
        // assert ((mm * i2l(m.v[0]))&INT32MASK) == INT32MASK;
        
        int[] tmp = new int[m.length+1];
        int[] x = new int[m.length+1];

        // R^2 mod m
        {
            // hold enough memory to calculate x mod m and R^2 mod m.
            int rl =  2*m.length+1;
            if (this.length > rl) rl = this.length;
            
            // yet another buffer for calulating this mod m an R^2 mod m.
            int [] xx = new int [rl];
            xx[m.length*2] = 1;
        
            mod0(xx,xx.length,m);
        
            if (compare0shifted(this.v,this.length,m.v,m.length,0) < 0) {
                
                // x is already stored modulo m.
                // mont(x,R^2 mod m)
                montmult0(x,xx,this.v,this.length,m,mm);
            }
            else {
                // x modulo m has to be calculated in advance.
                
                // store R^2 mod m in tmp.
                System.arraycopy(xx,0,tmp,0,m.length);
                // store this in xx, which is large enough.
                System.arraycopy(this.v,0,xx,0,this.length);
                
                mod0(xx,this.length,m);
                // mont(x mod m,R^2 mod m)
                montmult0(x,xx,tmp,m.length,m,mm);
            }
        } // let buffer xx go out of scope, it's not needed any more...
        
        // R mod m
        int [] a = new int [m.length+1];
        a[m.length] = 1;
        
        mod0(a,a.length,m);
        
        for (int i=el-1;i>=0;--i) {
        
            int nbits = i < el-1 ? 31 : highestBit(e[i]);
        
            for (int ibit=nbits;ibit>=0;--ibit) {
            
                montmult0(tmp,a,a,m.length,m,mm);
                
                // is bit i set in n?
                if ((e[i]&(1 << ibit)) != 0) {
                
                    montmult0(a,tmp,x,m.length,m,mm);
                }
                else {
                    int[] xxx=tmp; tmp=a; a=xxx;
                }
            }
        }
        
        x[0] = 1;
       
        montmult0(tmp,a,x,1,m,mm);
        
        return new BigInt((e[0]&1)==0 ? 1 : this.sign,tmp);
    }
}
