/***********************************************************
 * $Id: MGF1.java 216 2010-03-25 22:25:41Z wglas $
 * 
 * optional of the clazzes.org project
 * 
 * http://www.clazzes.org
 *
 * Created: 23.03.2010, wglas
 *
 ***********************************************************/

package org.clazzes.optional.sec.rsa;

import org.clazzes.optional.sec.Digest;

/**
 * The mask generation function no. 1 for the OAEP algorithm as defined in 
 * <a href="http://tools.ietf.org/html/rfc3447#appendix-B.2">Appendix B.2.1 of RFC 3447</a>.
 */
public class MGF1 implements MGF {
    
    private final Digest digest;

    public MGF1(Digest digest) {
        this.digest = digest;
    }
    
    /* (non-Javadoc)
     * @see org.clazzes.optional.sec.MGF#generateMask(byte[], int)
     */
    public byte[] generateMask(byte[] mgfSeed, int seedOff, int seedLen, byte[] ret, int off, int maskLen) {
        
        /*
           MGF1 is a Mask Generation Function based on a hash function.

           MGF1 (mgfSeed, maskLen)

           Options:
           Hash     hash function (hLen denotes the length in octets of the hash
                    function output)

           Input:
           mgfSeed  seed from which mask is generated, an octet string
           maskLen  intended length in octets of the mask, at most 2^32 hLen

           Output:
           mask     mask, an octet string of length maskLen

           Error:   "mask too long"

           Steps:

           1. If maskLen > 2^32 hLen, output "mask too long" and stop.

           2. Let T be the empty octet string.

           3. For counter from 0 to \ceil (maskLen / hLen) - 1, do the
              following:

              a. Convert counter to an octet string C of length 4 octets (see
                 Section 4.1):

                    C = I2OSP (counter, 4) .

              b. Concatenate the hash of the seed mgfSeed and C to the octet
                 string T:

                    T = T || Hash(mgfSeed || C) .

             4. Output the leading maskLen octets of T as the octet string mask.

         */

        int hLen = this.digest.getDigestSize();
        int outoff = 0;
        
        byte [] tmp = new byte[hLen];
        int counter = 0;
        
        while (outoff < maskLen) {
        
            // MSB first, as per Section 4.1 :-/
            tmp[0] = (byte)(counter>>>24);
            tmp[1] = (byte)(counter>>>16);
            tmp[2] = (byte)(counter>>>8);
            tmp[3] = (byte) counter;
            
            this.digest.update(mgfSeed,seedOff,seedLen);
            this.digest.update(tmp,0,4);
            this.digest.doFinal(tmp,0);
            
            int nout = (hLen > maskLen-outoff) ? maskLen-outoff : hLen;
            System.arraycopy(tmp,0,ret,off+outoff,nout);
            outoff += nout;
            ++counter;
        }
        
        return ret;
    }

    /* (non-Javadoc)
     * @see org.clazzes.optional.sec.MGF#getUnderlyingDigest()
     */
    public Digest getUnderlyingDigest() {
        
        return this.digest;
    }

}
