/***********************************************************
 * $Id: OAEP.java 253 2010-04-16 08:07:16Z util $
 * 
 * optional of the clazzes.org project
 * 
 * http://www.clazzes.org
 *
 * Created: 23.03.2010, wglas
 *
 ***********************************************************/

package org.clazzes.optional.sec.rsa;

import org.clazzes.optional.lang.BigInt;
import org.clazzes.optional.sec.Digest;
import org.clazzes.optional.sec.RandomGenerator;

/**
 * OAEP algorithm as defined in 
 * <a href="http://tools.ietf.org/html/rfc3447#section-7.1">Section 7.1 of RFC 3447</a>.
 */
public class OAEP {

    private final Digest digest;
    private final MGF mgf;
    private final RandomGenerator randomGenerator;
    
    public OAEP(Digest digest, MGF mgf, RandomGenerator randomGenerator) {
        this.digest = digest;
        this.mgf = mgf;
        this.randomGenerator = randomGenerator;
    }

    /**
     * Prepare the given input data for RSA encryption. 
     * 
     * @param in The input buffer
     * @param inOff The offset of the data to be encrypted.
     * @param inLen The length of the data to be encrypted.
     * @param label The encryption label used as additional seed material.
     * @param out The output buffer. The length of the output buffer is the
     *            size of the RSA modulus used.
     */
    public void encrypt(byte[] in, int inOff, int inLen, byte[] label, byte[] out) {
        
        /*
            1. Length checking:

               a. If the length of L is greater than the input limitation for the
                  hash function (2^61 - 1 octets for SHA-1), output "label too
                  long" and stop.

               b. If mLen > k - 2hLen - 2, output "message too long" and stop.

            2. EME-OAEP encoding (see Figure 1 below):

               a. If the label L is not provided, let L be the empty string. Let
                  lHash = Hash(L), an octet string of length hLen (see the note
                  below).

               b. Generate an octet string PS consisting of k - mLen - 2hLen - 2
                  zero octets.  The length of PS may be zero.

               c. Concatenate lHash, PS, a single octet with hexadecimal value
                  0x01, and the message M to form a data block DB of length k -
                  hLen - 1 octets as

                     DB = lHash || PS || 0x01 || M.

               d. Generate a random octet string seed of length hLen.

               e. Let dbMask = MGF(seed, k - hLen - 1).

               f. Let maskedDB = DB \xor dbMask.

               g. Let seedMask = MGF(maskedDB, hLen).

               h. Let maskedSeed = seed \xor seedMask.

               i. Concatenate a single octet with hexadecimal value 0x00,
                  maskedSeed, and maskedDB to form an encoded message EM of
                  length k octets as

                     EM = 0x00 || maskedSeed || maskedDB.
         */
        
        int hLen = this.digest.getDigestSize();
        
        if (inLen > out.length - 2*hLen - 2)
            throw new SecurityException("Message size exceed permissible length ["+(out.length - 2*hLen - 2)+"].");
        
        // layout: | lHash || PS || 0x01 || M |
        //         +------------DB------------+
        byte [] db = new byte[out.length - hLen - 1];
        
        // calculate lHash
        if (label != null)
            digest.update(label,0,label.length);
        
        digest.doFinal(db,0);
        
        // lone 0x01 octet
        db[db.length-inLen-1] = 0x01;
        
        // append plain text
        System.arraycopy(in,inOff,db,db.length-inLen,inLen);
        
        // generate seed.
        byte [] seed = new byte[hLen]; 
        this.randomGenerator.nextBytes(seed);
        
        out[0] = 0x00; 
        
        // dbMask
        this.mgf.generateMask(seed,0,seed.length,out,hLen+1,out.length-hLen-1);
        
        // maskedDB
        for (int i=0;i<out.length-hLen-1;++i)
            out[hLen+1+i] ^= db[i];
        
        // seedMask
        this.mgf.generateMask(out,hLen+1,out.length-hLen-1,out,1,hLen);
        
        // maskedSeed
        for (int i=0;i<hLen;++i)
            out[i+1] ^= seed[i];
    }
    
    /**
     * Pad an encrypt the given input data using RSA encryption. 
     * 
     * @param in The input buffer
     * @param inOff The offset of the data to be encrypted.
     * @param inLen The length of the data to be encrypted.
     * @param pubKey The RSA public key used to encrypt.
     * @param label The encryption label used as additional seed material.
     */
    public byte[] encryptRSA(byte[] in, int inOff, int inLen,
            RSAPubKey pubKey, byte[] label) {
        
        byte [] out = new byte[pubKey.getModulus().getByteLength()];
        encrypt(in, inOff, inLen, label, out);
        
        BigInt out_bi = new BigInt(1,out);
        
        BigInt enc_bi = pubKey.encryptPlain(out_bi);
        return enc_bi.getMsbMagnitude();
    }
    
    /**
     * Prepare the given input data for RSA encryption. 
     * 
     * @param in The input buffer, which not longer than <code>keyLen</code>.
     * @param keyLen The length of the RSA modulus. THe length of <code>in</code>
     *               may be less than <code>keyLen</code>, because <code>in</code>
     *               is the magnitude of a big integer, which might have contained
     *               leading zero bytes. Hence, if <code>in</code> is shorter than
     *                <code>keyLen</code>, the missing parts are interpreted as
     *               leading zero bytes.
     * @param label The encryption label used as additional seed material.
     * @param out The output buffer. The length of the output buffer must be at least
     *            the size of the RSA modulus used.
     * @param outOff The offset at which the result is stored in <code>out</code>.
     * 
     * @return The number of bytes stored to <code>out</code>.
     */
    public int decrypt(byte[] in, int keyLen, byte[] label, byte[] out, int outOff) {

        /*
            3. EME-OAEP decoding:

              a. If the label L is not provided, let L be the empty string. Let
                 lHash = Hash(L), an octet string of length hLen (see the note
                 in Section 7.1.1).

              b. Separate the encoded message EM into a single octet Y, an octet
                 string maskedSeed of length hLen, and an octet string maskedDB
                 of length k - hLen - 1 as

                    EM = Y || maskedSeed || maskedDB.

              c. Let seedMask = MGF(maskedDB, hLen).

              d. Let seed = maskedSeed \xor seedMask.

              e. Let dbMask = MGF(seed, k - hLen - 1).

              f. Let DB = maskedDB \xor dbMask.

              g. Separate DB into an octet string lHash' of length hLen, a
                 (possibly empty) padding string PS consisting of octets with
                 hexadecimal value 0x00, and a message M as

                    DB = lHash' || PS || 0x01 || M.

                 If there is no octet with hexadecimal value 0x01 to separate PS
                 from M, if lHash does not equal lHash', or if Y is nonzero,
                 output "decryption error" and stop.  (See the note below.)

         */
        
        if (in.length > keyLen)
            throw new SecurityException("Input data is longer than key length.");
        
        int hLen = this.digest.getDigestSize();
        
        // layout: | lHash || seedMask || dbMask |
        //         +------------DB------------+
        byte [] db = new byte[keyLen - 1 + hLen];
        
        // calculate lHash
        if (label != null)
            this.digest.update(label,0,label.length);
        
        this.digest.doFinal(db,0);
        
        // seedMask
        this.mgf.generateMask(in,in.length-(keyLen-hLen-1),keyLen-hLen-1,db,hLen,hLen);
        
        // calculate seed at position hLen...hLen+hLen-1 in array db.
        int i= in.length < keyLen-1 ? keyLen-1-in.length : 0;
        int inOff = in.length < keyLen ? 0 : 1; 
        
        while (i<hLen) {
            
            db[hLen+i] ^= in[inOff+i];
            ++i;
        }
        
        // dbMask
        this.mgf.generateMask(db,hLen,hLen,db,hLen+hLen,keyLen-hLen-1);
        
        for (i=0;i<keyLen-hLen-1;++i) {
            
            db[2*hLen+i] ^= in[in.length - (keyLen-hLen-1) + i];
        }
        
        if (inOff > 1 && in[0] != 0x00)
            throw new SecurityException("Leading byte is not zero.");
        
        // check lHash and lHash'
        boolean ok = true;
        
        for (i=0;i<hLen;++i) {
            if ((db[i] ^ db[2*hLen+i]) != 0) ok = false;
        }
        
        // check zero bytes
        for (;i<keyLen-hLen-1;++i) {
            
            if (db[2*hLen+i] == 0x00) continue;
            if (db[2*hLen+i] == 0x01) break;
            ok = false;
        }
        
        if (i>=keyLen-hLen-1) ok = false;
        
        if (!ok)
            throw new SecurityException("Malformed padded plaintext.");
        
        ++i;
        int n = db.length - (2*hLen + i);
        
        System.arraycopy(db,2*hLen+i,out,outOff,n);
        
        return n;
    }
    
    /**
     * Decrypt and unpad the given input data using RSA decryption. 
     * 
     * @param in The input buffer, which not longer than <code>keyLen</code>.
     * @param privKey The RAS private key used for the plain RSA operation.
     * @param label The encryption label used as additional seed material.
     * @param out The output buffer. The length of the output buffer must be at least
     *            the size of the RSA modulus used.
     * @param outOff The offset at which the result is stored in <code>out</code>.
     * 
     * @return The number of bytes stored to <code>out</code>.
     */
  //#ifndef JAVAME
    public int  decryptRSA(byte[] in, RSAPrivKey privKey, byte[] label, byte[] out, int outOff) {
        
        BigInt in_bi = new BigInt(1,in);
        
        BigInt dec_bi = privKey.decryptPlain(in_bi);

        return this.decrypt(dec_bi.getMsbMagnitude(),privKey.getModulus().getByteLength(),label,out,outOff);
    }
  //#endif
}