/***********************************************************
 * $Id: RSACrtPrivKey.java 253 2010-04-16 08:07:16Z util $
 * 
 * optional of the clazzes.org project
 * 
 * http://www.clazzes.org
 *
 * Created: 05.04.2010, wglas
 *
 ***********************************************************/

package org.clazzes.optional.sec.rsa;

import java.io.DataOutputStream;
import java.io.IOException;

import org.clazzes.optional.lang.BigInt;

/**
 * An RSA private key with a passphrase-encrypted storage possibility. 
 * 
 * This specialized class does make use of the Chinese Remainder Theorem and hence stores
 * the factorization of the modulus and other values alongside with the private exponent.
 */
public class RSACrtPrivKey extends RSAPrivKey {
    
    private final BigInt p;
    private final BigInt q;
    private final BigInt dp;
    private final BigInt dq;
    private final BigInt qi;
    
    /**
     * @param modulus The modulus of the RSA key.
     * @param pubExponent The public exponent.
     * @param privExponent The private exponent.
     * @param p The larger prime of the factorization of the modulus.
     * @param q The smaller prime of the factorization of the modulus.
     * @param dp <code>privExponent mod (p-1)</code>
     * @param dq <code>privExponent mod (q-1)</code>
     * @param qi The modular inverse of <code>q</code> w.r.t. <code>p</code>. 
     *           (<code>(q * qi) mod p == 1 mod p</code>)
     */
    public RSACrtPrivKey(BigInt modulus,
            BigInt pubExponent,
            BigInt privExponent,
            BigInt p, BigInt q, BigInt dp, BigInt dq, BigInt qi) {
        super(modulus, pubExponent, privExponent);
        this.p = p;
        this.q = q;
        this.dp = dp;
        this.dq = dq;
        this.qi = qi;
    }

    public BigInt decryptPlain(BigInt x) {
        
        BigInt decp = x.powMod(this.dp,this.p);
        BigInt decq = x.powMod(this.dq,this.q);

        BigInt delta = decp.sub(decq);
        if (delta.getSign() < 0) {
            delta = delta.add(this.p);
        }

        return delta.mulMod(this.qi,this.p).mul(this.q).add(decq).mod(this.getModulus());
    }

    protected void encodeEncryptedParts(DataOutputStream dos)
            throws IOException {
        
        super.encodeEncryptedParts(dos);
        this.p.writeData(dos);
        this.q.writeData(dos);
        this.dp.writeData(dos);
        this.dq.writeData(dos);
        this.qi.writeData(dos);
    }

    
}
