/***********************************************************
 * $Id: RSAPrivKey.java 253 2010-04-16 08:07:16Z util $
 * 
 * optional of the clazzes.org project
 * 
 * http://www.clazzes.org
 *
 * Created: 03.04.2010, wglas
 *
 ***********************************************************/

package org.clazzes.optional.sec.rsa;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.clazzes.optional.lang.BigInt;
import org.clazzes.optional.sec.AESCipher;
import org.clazzes.optional.sec.GCMCipher;
import org.clazzes.optional.sec.RandomGenerator;
import org.clazzes.optional.sec.SHA256Digest;

/**
 * An RSA private key with a passphrase-encrypted storage possibility. 
 * 
 * This class does not make use of the Chinese Remainder Theorem and hence stores
 * the private exponent internally without the factorization of the modulus.
 */
public class RSAPrivKey {
    
    private final BigInt privExponent;
    private final BigInt modulus;
    private final BigInt pubExponent;
   
    // old serialization format with legacy BigInt serialization.
    private static final int SERIALIZATION_VERSION_LEGACY = 1;
    
    // new serialization format with TLS-like BigInt serialization.
    private static final int SERIALIZATION_VERSION = 2;
   
    /**
     * Construct an RSA private key from the given numbers.
     * 
     * @param modulus The modulus of operation.
     * @param pubExponent The public exponent.
     * @param privExponent The private exponent.
     */
    public RSAPrivKey(BigInt modulus, BigInt pubExponent, BigInt privExponent) {
      
        this.modulus = modulus;
        this.pubExponent = pubExponent;
        this.privExponent = privExponent;
    }
    
    /**
     * @return The modulus of the RAS public key.
     */
    public BigInt getModulus() {
        return this.modulus;
    }

    /**
     * @return The public exponent of the RSA key.
     */
    public BigInt getPubExponent() {
        return this.pubExponent;
    }
    
    /**
     * @return The associated public key.
     */
    public RSAPubKey getPubKey() {
        return new RSAPubKey(this.modulus,this.pubExponent);
    }

    /**
     * Decode an unsigned BigInt from a DataInputStream.
     * @param dis The DataInputStream to read from.
     * @return The positive number decoded from the stream.
     * @throws IOException Upon write errors.
     */
    protected static BigInt decodeBigInt(DataInputStream dis, boolean legacy) throws IOException {
        
        if (legacy) {
        
            int l = dis.readInt();
        
            if (l < 1 || l > 128)
                throw new IOException("Invalid length of RSA part.");

            int [] dwords = new int[l];
        
            for (int i=0;i<l;++i)
                dwords[i] = dis.readInt();
            
            return new BigInt(1,dwords);
        }
        else {
            return BigInt.readDataInstance(dis);
        }
    }
    
    /**
     * Encode the encrypted parts. This function may be overwritten by more
     * sophisticated implementations like implementations, which make use of
     * the Chinese Remainder Theorem.
     *  
     * @param dos The data output stream, which will be encrypted.
     * @throws IOException
     */
    protected void encodeEncryptedParts(DataOutputStream dos) throws IOException {
        
        this.modulus.writeData(dos);
        this.pubExponent.writeData(dos);
        this.privExponent.writeData(dos);
    }
    
    /**
     * Write the encrypted private key to an output stream.
     * 
     * @param os The target output stream.
     * @param passPhrase A pass phrase. If the pass phrase is a string it is
     *                   highly recommended to use the UTF-8 representation of
     *                   the pass phrase.
     * @param random A random number generator used to generate salt bytes.
     * @throws IOException Upon write errors.
     */
    public void encode(OutputStream os, byte [] passPhrase, RandomGenerator random) throws IOException {
        
        if (passPhrase == null || passPhrase.length < 8)
            throw new SecurityException("RSA pass phrase is whorter than 8 bytes.");
        
        SHA256Digest sha256 = new SHA256Digest();
        
        byte [] salt = new byte[32];
        random.nextBytes(salt);
        sha256.update(salt,0,salt.length);
        sha256.update(passPhrase,0,passPhrase.length);
        byte [] key = new byte[32];
        sha256.doFinal(key,0);
        
        byte [] aad = new byte[8];
       
        long now = System.currentTimeMillis();
        
        aad[0] = (byte)(now >>> 56);
        aad[1] = (byte)(now >>> 48);
        aad[2] = (byte)(now >>> 40);
        aad[3] = (byte)(now >>> 32);
        aad[4] = (byte)(now >>> 24);
        aad[5] = (byte)(now >>> 16);
        aad[6] = (byte)(now >>> 8);
        aad[7] = (byte)(now);
        
        byte [] nonce = new byte[12];
        random.nextBytes(nonce);
        
        GCMCipher cipher = new GCMCipher(new AESCipher());
        cipher.getUnderlyingCipher().init(key,true);
        cipher.init(aad,112,nonce,true);
        
        DataOutputStream dos = new DataOutputStream(os);
        
        dos.writeInt(SERIALIZATION_VERSION);   
        dos.write(salt);
        dos.write(aad);
        dos.write(nonce);
        
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        {
            DataOutputStream dose = new DataOutputStream(bos);
            
            this.encodeEncryptedParts(dose);
            dose.flush();
        }
        
        byte [] plain = bos.toByteArray();
        byte [] encrypted = new byte[cipher.getOutputSize(plain.length)];
        
        int nenc = cipher.processBytes(plain,0,plain.length,encrypted,0);
        nenc += cipher.doFinal(encrypted,nenc);
        
        dos.writeInt(nenc);
        dos.write(encrypted,0,nenc);
        dos.flush();
    }
    
    /**
     * Write the encrypted private key to an output stream.
     * 
     * @param os The target output stream.
     * @param passPhrase A pass phrase. If the pass phrase is a string it is
     *                   highly recommended to use the UTF-8 representation of
     *                   the pass phrase.
     * @throws IOException Upon read errors.
     */
    public static RSAPrivKey decode(InputStream is, byte [] passPhrase) throws IOException {
        
        if (passPhrase == null || passPhrase.length < 8)
            throw new SecurityException("RSA pass phrase is whorter than 8 bytes.");
        
        SHA256Digest sha256 = new SHA256Digest();
        
        byte [] salt = new byte[32];
        
        byte [] aad = new byte[8];
        byte [] nonce = new byte[12];
              
        DataInputStream dis = new DataInputStream(is);
        
        int v = dis.readInt();
        if (v != SERIALIZATION_VERSION && v != SERIALIZATION_VERSION_LEGACY)
            throw new IOException("Invalid serialization format ["+v+"] specified.");
        
        dis.read(salt);
        dis.read(aad);
        dis.read(nonce);

        sha256.update(salt,0,salt.length);
        sha256.update(passPhrase,0,passPhrase.length);
        byte [] key = new byte[32];
        sha256.doFinal(key,0);
         
        GCMCipher cipher = new GCMCipher(new AESCipher());
        cipher.getUnderlyingCipher().init(key,true);
        cipher.init(aad,112,nonce,false);
                
        int nenc = dis.readInt();
        
        if (nenc < 0 || nenc > 16384)
            throw new IOException("Cipher text for RSA key exceeds 16k.");
        
        byte [] encrypted = new byte[nenc];
        dis.read(encrypted);

        byte [] plain = new byte[cipher.getOutputSize(encrypted.length)];
        
        int nplain = cipher.processBytes(encrypted,0,encrypted.length,plain,0);
        nplain += cipher.doFinal(plain,nplain);

        ByteArrayInputStream bis = new ByteArrayInputStream(plain,0,nplain);
        DataInputStream bise = new DataInputStream(bis);
        
        boolean legacy = (v == SERIALIZATION_VERSION_LEGACY);
        
        BigInt modulus = decodeBigInt(bise,legacy);
        BigInt pubExponent = decodeBigInt(bise,legacy);
        BigInt privExponent =  decodeBigInt(bise,legacy);
        
        if (bise.available() > 0) {
       
            BigInt p = decodeBigInt(bise,legacy);
            BigInt q = decodeBigInt(bise,legacy);
            BigInt dp = decodeBigInt(bise,legacy);
            BigInt dq = decodeBigInt(bise,legacy);
            BigInt qi = decodeBigInt(bise,legacy);

            if (bise.available() > 0)
                throw new IOException("Extra information at end of encrypted RSA information.");
            
            return new RSACrtPrivKey(modulus,pubExponent,privExponent,p,q,dp,dq,qi);
        }
        
        return new RSAPrivKey(modulus,pubExponent,privExponent);
     }
    
   /**
     * Calculates the plain decrpyt operation
     * <code>x^privExp</code>. This operation may additionally
     * be used on a padded message digest to sign data.
     * 
     * @param x The encrypted text.
     * @return The decrypted plain text, which is usually subject to
     *         an appropriate padding scheme.
     */
    public BigInt decryptPlain(BigInt x) {
        return x.powMod(this.privExponent,this.modulus);
    }
   
}
